import logging
import multiprocessing
import os
import shutil

import subprocess

from ..utils.exceptions import OasisException
from ..utils.log import oasis_log
from .bash import genbash


@oasis_log()
def run(analysis_settings,
        number_of_processes=-1,
        set_alloc_rule_gul=None,
        set_alloc_rule_il=None,
        set_alloc_rule_ri=None,
        gul_legacy_stream=False,
        run_debug=False,
        custom_gulcalc_cmd=None,
        filename='run_ktools.sh',
        **kwargs
):
    # TODO: need clearer responsibility between runner.py and manager.py
    #  ie: why is cpu count and custom_gulcalc_cmd checks here and not in manager
    #      as manager does a lot of alike env check.
    if number_of_processes == -1:
        number_of_processes = multiprocessing.cpu_count()

    # If `given_gulcalc_cmd` is set then always run as a complex model
    # and raise an exception when not found in PATH
    if custom_gulcalc_cmd:
        if not shutil.which(custom_gulcalc_cmd):
            raise OasisException(
                'Run error: Custom Gulcalc command "{}" explicitly set but not found in path.'.format(custom_gulcalc_cmd)
            )
    # when not set then fallback to previous behaviour:
    # Check if a custom binary `<supplier>_<model>_gulcalc` exists in PATH
    else:
        inferred_gulcalc_cmd = "{}_{}_gulcalc".format(
            analysis_settings.get('module_supplier_id'),
            analysis_settings.get('model_version_id'))
        if shutil.which(inferred_gulcalc_cmd):
            custom_gulcalc_cmd = inferred_gulcalc_cmd

    # TODO: should be integrated into bash.py
    if custom_gulcalc_cmd:
        def custom_get_getmodel_cmd(
            number_of_samples,
            gul_threshold,
            use_random_number_file,
            coverage_output,
            item_output,
            process_id,
            max_process_id,
            gul_alloc_rule,
            stderr_guard,
            **kwargs
        ):

            cmd = "{} -e {} {} -a {} -p {}".format(
                custom_gulcalc_cmd,
                process_id,
                max_process_id,
                os.path.abspath("analysis_settings.json"),
                "input")
            if gul_legacy_stream and coverage_output != '':    
                cmd = '{} -c {}'.format(cmd, coverage_output)
            if item_output != '':
                cmd = '{} -i {}'.format(cmd, item_output)
            if stderr_guard:
                cmd = '({}) 2>> log/gul_stderror.err'.format(cmd)

            return cmd
    else:
        custom_get_getmodel_cmd = None

    genbash(
        number_of_processes,
        analysis_settings,
        gul_alloc_rule=set_alloc_rule_gul,
        il_alloc_rule=set_alloc_rule_il,
        ri_alloc_rule=set_alloc_rule_ri,
        gul_legacy_stream=gul_legacy_stream,
        bash_trace=run_debug,
        filename=filename,
        _get_getmodel_cmd=custom_get_getmodel_cmd,
        **kwargs,
    )

    bash_trace = subprocess.check_output(['bash', filename])
    logging.info(bash_trace.decode('utf-8'))



"""
Module responsible for reading and writing FM stream.

Some detail on the data layout.
event id fron the stream can be:
 - negative for special values like the mean or the standard deviation.
 - positive for normal sample id
in the loss array:
 - sidx -3 correspond to index 0
 - sidx -2 is filtered
 - all other sidx egual their index
 - value are arranged so the index in the array mostly correspond to the sample id,
   the special values -1 is place at the end therefor.
   ie: [sidx_-3, sidx_1, sidx_2, ...sidx_len_sample, sidx_-1]
   here we use python ability to understand negative indexes to equate sidx and index simplifying index management


"""

from .common import float_equal_precision

import sys
from numba import njit
import numpy as np
import selectors
from select import select
import logging
logger = logging.getLogger(__name__)


#gul_header = np.int32(1 << 24).tobytes()
fm_header = np.int32(1 | 2 << 24).tobytes()


buff_size = 65536
event_agg_dtype = np.dtype([('event_id', 'i4'), ('agg_id', 'i4')])
sidx_loss_dtype = np.dtype([('sidx', 'i4'), ('loss', 'f4')])
number_size = 8
nb_number = buff_size // number_size

EXTRA_VALUES = 2


@njit(cache=True)
def stream_to_loss_table(event_agg, sidx_loss, valid_buf, cursor, event_id, agg_id, loss_index, nodes_array, losses, index, computes):
    """valid len must be divisible par 2*4 bytes
    cursor, event_id, agg_id, loss_index, yield_event
    """
    valid_len = valid_buf // number_size

    if agg_id:
        last_event_id = event_id
        node = nodes_array[agg_id]
        index[node['loss']: node['loss'] + node['layer_len']] = loss_index
        computes[loss_index] = agg_id
    else:
        last_event_id = event_id

    while cursor < valid_len:
        if agg_id:
            sidx, loss = sidx_loss[cursor]['sidx'], sidx_loss[cursor]['loss']
            cursor += 1

            if sidx:
                if sidx == -2:
                    continue
                elif sidx == -3:
                    sidx = 0
                losses[loss_index, sidx] = 0 if np.isnan(loss) else loss
            else:
                agg_id = 0
                loss_index += 1
        else:
            event_id, agg_id = event_agg[cursor]['event_id'], event_agg[cursor]['agg_id']
            cursor += 1
            if event_id != last_event_id:
                if last_event_id:
                    return cursor - 1, last_event_id, 0, loss_index, 1
                else:
                    last_event_id = event_id
            node = nodes_array[agg_id]
            index[node['loss']: node['loss'] + node['layer_len']] = loss_index
            computes[loss_index] = agg_id
            losses[loss_index].fill(0)

    return cursor, event_id, agg_id, loss_index, 0


def read_stream_header(stream_obj):
    stream_type = stream_obj.read(4)
    len_sample = np.frombuffer(stream_obj.read(4), dtype=np.int32)[0]
    return stream_type, len_sample


def read_event(stream, nodes_array, losses, index, computes, main_selector,
               stream_selector, mv, event_agg, sidx_loss, cursor, valid_buf):
    event_id = 0
    agg_id = 0
    loss_index = 0

    while True:
        if valid_buf < buff_size:
            len_read = stream.readinto1(mv[valid_buf:])
            valid_buf += len_read

            if len_read == 0:
                stream_selector.close()
                main_selector.unregister(stream)
                if event_id:
                    if agg_id:
                        loss_index += 1
                    return event_id, loss_index, cursor, valid_buf

                break
        cursor, event_id, agg_id, loss_index, yield_event = stream_to_loss_table(event_agg, sidx_loss, valid_buf,
                                                                                 cursor,
                                                                                 event_id, agg_id, loss_index,
                                                                                 nodes_array, losses, index, computes)

        if yield_event:
            if number_size * cursor == valid_buf:
                valid_buf = 0
            return event_id, loss_index, cursor, valid_buf
        else:
            cursor_buf = number_size * cursor
            mv[:valid_buf - cursor_buf] = mv[cursor_buf: valid_buf]
            valid_buf -= cursor_buf
            cursor = 0
            stream_selector.select()


def register_streams_in(selector_class, streams_in):
    """
    Data from input process is generally sent by event block, meaning once a stream receive data, the complete event is
    going to be sent in a short amount of time.
    Therefore, we can focus on each stream one by one using their specific selector 'stream_selector'.

    """
    main_selector = selector_class()
    stream_data = []
    for stream_in in streams_in:
        mv = memoryview(bytearray(buff_size))

        event_agg = np.ndarray(nb_number, buffer=mv, dtype=event_agg_dtype)
        sidx_loss = np.ndarray(nb_number, buffer=mv, dtype=sidx_loss_dtype)

        stream_selector = selector_class()
        stream_selector.register(stream_in, selectors.EVENT_READ)
        data = {'mv': mv,
                'event_agg': event_agg,
                'sidx_loss': sidx_loss,
                'cursor': 0,
                'valid_buf': 0,
                'stream_selector': stream_selector
               }
        stream_data.append(data)
        main_selector.register(stream_in, selectors.EVENT_READ, data)
    return main_selector, stream_data


def read_streams(streams_in, nodes_array, losses, index, computes):
    try:
        main_selector, stream_data = register_streams_in(selectors.DefaultSelector, streams_in)
        logger.debug("Streams read with DefaultSelector")
    except PermissionError: # Fall back option if stream_in contain regular files
        main_selector, stream_data = register_streams_in(selectors.SelectSelector, streams_in)
        logger.debug("Streams read with SelectSelector")
    try:
        while main_selector.get_map():
            for sKey, _ in main_selector.select():
                event = read_event(sKey.fileobj, nodes_array, losses, index, computes, main_selector, **sKey.data)
                if event:
                    event_id, loss_index, cursor, valid_buf = event
                    sKey.data['cursor'] = cursor
                    sKey.data['valid_buf'] = valid_buf
                    logger.debug(f'event_id: {event_id}, loss_index :{loss_index}')
                    yield event_id, loss_index

        # Stream is read, we need to check if there is remaining event to be parsed
        for data in stream_data:
            if data['cursor'] < data['valid_buf']:
                event_agg = data['event_agg']
                sidx_loss = data['sidx_loss']
                cursor = data['cursor']
                valid_buf = data['valid_buf']
                yield_event = True
                while yield_event:
                    cursor, event_id, agg_id, loss_index, yield_event = stream_to_loss_table(event_agg,
                                                                                             sidx_loss,
                                                                                             valid_buf,
                                                                                             cursor,
                                                                                             0, 0,
                                                                                             0,
                                                                                             nodes_array, losses, index,
                                                                                             computes)

                    if event_id:
                        if not yield_event and agg_id:
                            loss_index += 1
                        logger.debug(f'event_id: {event_id}, loss_index :{loss_index}')
                        yield event_id, loss_index

    finally:
        main_selector.close()


@njit(cache=True)
def load_event(event_agg, sidx_loss, event_id, nodes_array, losses, loss_indexes, computes, output_array, compute_i, i_layer, i_index, nb_values):
    cursor = 0
    top_sidx_range = losses.shape[1] - EXTRA_VALUES + 1

    while computes[compute_i]:
        node = nodes_array[computes[compute_i]]
        for layer in range(i_layer, node['layer_len']):
            output_id = output_array[node['output_ids']+layer]
            if output_id:  # if output is not in xref output_id is 0
                loss_index = loss_indexes[node['ba']+layer]
                #print(computes[compute_i], output_id, loss_index, losses[loss_index])
                while cursor < nb_values:
                    if i_index == 0:
                        event_agg[cursor]['event_id'], event_agg[cursor]['agg_id'] = event_id, output_id
                        cursor += 1
                        i_index = -3

                    elif i_index == -3:
                        sidx_loss[cursor]['sidx'], sidx_loss[cursor]['loss'] = -3, losses[loss_index, 0]
                        cursor += 1
                        i_index = -1
                    elif i_index == -1:
                        sidx_loss[cursor]['sidx'], sidx_loss[cursor]['loss'] = -1, losses[loss_index, -1]
                        cursor += 1
                        i_index = 1

                    elif i_index == top_sidx_range:
                        sidx_loss[cursor]['sidx'], sidx_loss[cursor]['loss'] = 0, 0
                        cursor += 1
                        i_index = 0
                        i_layer = 0
                        break

                    else:
                        if losses[loss_index, i_index] > float_equal_precision:
                            sidx_loss[cursor]['sidx'], sidx_loss[cursor]['loss'] = i_index, losses[loss_index, i_index]
                            cursor += 1
                        i_index += 1
                else:
                    return cursor * number_size, compute_i, layer, i_index
        compute_i += 1

    return cursor * number_size, compute_i, 0, i_index


class EventWriter:
    def __init__(self, files_out, nodes_array, output_array, losses, loss_indexes, computes, len_sample):
        self.files_out = files_out
        self.nodes_array = nodes_array
        self.losses = losses
        self.loss_indexes = loss_indexes
        self.computes = computes
        self.output_array = output_array
        self.nb_output_items = output_array.shape[0]

        self.len_sample = len_sample
        self.loss_shape_1 = len_sample + EXTRA_VALUES # all normal sidx plus the extra values

        self.mv = memoryview(bytearray(nb_number * number_size))

        self.event_agg = np.ndarray(nb_number, buffer=self.mv, dtype=event_agg_dtype)
        self.sidx_loss = np.ndarray(nb_number, buffer=self.mv, dtype=sidx_loss_dtype)

    def __enter__(self):
        if self.files_out is None:
            self.stream_out = sys.stdout.buffer
        else:
            self.stream_out = open(self.files_out, 'wb')
        # prepare output buffer
        self.stream_out.write(fm_header)
        self.stream_out.write(np.int32(self.len_sample).tobytes())
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.files_out:
            self.stream_out.close()

    def write(self, event_id, compute_i):
        i_index= 0
        i_layer = 0
        stream_out = [self.stream_out]
        while self.computes[compute_i]:
            cursor, compute_i, i_layer, i_index = load_event(self.event_agg,
                                                              self.sidx_loss,
                                                              event_id,
                                                              self.nodes_array,
                                                              self.losses,
                                                              self.loss_indexes,
                                                              self.computes,
                                                              self.output_array,
                                                              compute_i,
                                                              i_layer,
                                                              i_index, nb_number)

            _, writable, exceptional = select([], stream_out, stream_out)
            if exceptional:
                raise IOError(f'error with input stream, {exceptional}')
            writable[0].write(self.mv[:cursor])
        return compute_i


@njit(cache=True)
def get_compute_end(computes, compute_i):
    while computes[compute_i]:
        compute_i += 1
    return compute_i


class EventWriterOrderedOutput(EventWriter):
    def write(self, event_id, compute_i):
        compute_end = get_compute_end(self.computes, compute_i)
        self.computes[compute_i: compute_end] = np.sort(self.computes[compute_i: compute_end])
        return super().write(event_id, compute_i)



from ..utils.exceptions import OasisException

from .admin import AdminCmd
from .config import ConfigCmd
from .exposure import ExposureCmd
from .model import ModelCmd
from .command import OasisBaseCommand
from .test import TestCmd
from .version import VersionCmd
from .api import ApiCmd


class RootCmd(OasisBaseCommand):
    """
    Tool for manageing oasislmf models
    """
    sub_commands = {
        'admin': AdminCmd,
        'test': TestCmd,
        'version': VersionCmd,
        'model': ModelCmd,
        'exposure': ExposureCmd,
        'api': ApiCmd,
        'config': ConfigCmd
    }

    def run(self, args=None):
        """
        Runs the command passing in the parsed arguments. If an ``OasisException`` is
        raised the exception is caught, the error is logged and the process exits with
        an error code of 1.

        :param args: The arguments to run the command with. If ``None`` the arguments
            are gathered from the argument parser. This is automatically set when calling
            sub commands and in most cases should not be set for the root command.
        :type args: Namespace

        :return: The status code of the action (0 on success)
        """
        try:
            return super(OasisBaseCommand, self).run(args=args)
        except OasisException as e:
            if self.log_verbose:
                # Log with traceback
                self.logger.exception(str(e))
            else:
                self.logger.error(str(e))
            return 1



__all__ = [
    'InputValues',
]

import io
import os
import json
import logging

from ..utils.defaults import get_config_profile
from ..utils.exceptions import OasisException
from argparse import ArgumentTypeError


class InputValues(object):
    """
    Helper class for accessing the input values from either
    the command line or the configuration file.

    internal_update

    """
    def __init__(self, args, update_keys=True):
        self.logger = logging.getLogger()

        self.args = args
        self.config = {}
        self.config_fp = self.get('config', is_path=True)
        self.config_mapping = get_config_profile()

        if self.config_fp is not None:
            self.config = self.load_config_file()
            self.config_dir = os.path.dirname(self.config_fp)

        self.obsolete_keys = set(self.config) & set(self.config_mapping)

        self.list_obsolete_keys()
        if update_keys:
            self.update_config_keys()

    def list_obsolete_keys(self, fix_warning=True):
        if self.obsolete_keys:
            self.logger.warning('Depricated key(s) in MDK config:')
            for k in self.obsolete_keys:
                self.logger.warning('   {} : {}'.format(
                    k,
                    self.config_mapping[k],
                ))
            self.logger.warning('')
            if fix_warning:
                self.logger.warning('  To fix run: oasislmf config update'.format(self.config_fp))

    def update_config_keys(self):
        """
        If command line flags change between package versions, update them internally
        """
        for key in self.obsolete_keys:
            if not self.config_mapping[key]['deleted']:
                self.config[self.config_mapping[key]['updated_to']] = self.config[key]
            del self.config[key]

    def load_config_file(self):
        try:
            with io.open(self.config_fp, 'r', encoding='utf-8') as f:
                return json.load(f)
        except FileNotFoundError:
            raise OasisException('MDK config. file path {} provided does not exist'.format(self.config_fp))

    def write_config_file(self, config_fp):
        with io.open(config_fp, 'w', encoding='utf-8') as f:
            f.write(u'{}'.format(json.dumps(self.config, sort_keys=True, indent=4, ensure_ascii=False)))

    def confirm_action(self, question_str, no_confirm=False):
        self.logger.debug('Prompt user for confirmation')
        if no_confirm:
            return True
        try:
            check = str(input("%s (Y/N): " % question_str)).lower().strip()
            if check[:1] == 'y':
                return True
            elif check[:1] == 'n':
                return False
            else:
                self.logger.error('Enter "y" for Yes, "n" for No or Ctrl-C to exit.\n')
                return self.confirm_action(question_str)
        except KeyboardInterrupt:
            self.logger.error('\nexiting.')

    def get_env(self, name, dtype=None, default=None):
        env_var = os.getenv(f'OASIS_{name.upper()}', default=default)
        if dtype and env_var:
            return dtype(env_var)
        else:
            return env_var

    def has_env(self, name):
        return f'OASIS_{name.upper()}' in os.environ

    def get(self, name, default=None, required=False, is_path=False, dtype=None):
        """
        Gets the name parameter until found from:
          - the command line arguments.
          - the configuration file
          - the environment variable (put in uppercase)

        If it is not found then ``default`` is returned
        unless ``required`` is True in which case an ``OasisException`` is raised.

        :param name: The name of the parameter to lookup
        :type name: str

        :param default: The default value to return if the name is not
            found on the command line or in the configuration file.

        :param required: Flag whether the value is required, if so and
            the parameter is not found on the command line or in the
            configuration file an error is raised.
        :type required: bool

        :param is_path: Flag whether the value should be treated as a path and return an abspath,
            use config_dir as base dir if value comes from the config
        :type is_path: bool

        :param dtype: the class <type> of the value, if 'None' load as string by default
        :type: class

        :raise OasisException: If the value is not found and ``required``
            is True

        :return: The found value or the default
        """
        # Load order 0:  Get from CLI flag
        source = 'arg'
        value = getattr(self.args, name, None)

        # Load order 1: ENV override (intended for worker images)
        if str2bool(os.getenv('OASIS_ENV_OVERRIDE', default=False)) and self.has_env(name):
            source = 'env_override'
            value = self.get_env(name, dtype)

        # Load order 2: Get from config JSON
        if value is None:
            source = 'config'
            value = self.config.get(name)

        # Load order 3: Get from environment variable
        if value is None:
            source = 'env'
            value = self.get_env(name, dtype)

        if value is None and required:
            raise OasisException(
                'Required argument {} could not be found in the command args or the MDK config. file'.format(name)
            )

        # Load order 4: Get default value
        if value is None:
            source = 'default'
            value = default

        if is_path and value not in [None, ""] and not os.path.isabs(value):
            if source == 'config':
                value = os.path.join(self.config_dir, value)
            else:
                value = os.path.abspath(value)

        # Warn user of environment variable load
        if source == 'env_override':
            self.logger.warn(f'Warning - environment variable override: OASIS_{name.upper()}={value}')

        return value


def str2bool(v):
    """ Func type for loading strings to boolean values using argparse
        https://stackoverflow.com/a/43357954

        step_params:
            use: `'default': False, 'type': str2bool, 'const':True, 'nargs':'?', ...`

        CLI:
            oasislmf --some-flag
            oasislmf --some-flag <bool>

        oasislmf.json
        {"some_flag": true, ...}
    """
    if v is None:
        return v
    elif isinstance(v, bool):
        return v
    elif v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise ArgumentTypeError('Boolean value expected.')
